@@ -41,6 +41,7 @@
 # include "calc/longbits.h"
 # include "calc/byteswap.h"
 # include "calc/have_stdlib.h"
+# include "calc/attribute.h"
 #else
 # include <calc/decl.h>
 # include <calc/alloc.h>
@@ -48,6 +49,7 @@
 # include <calc/longbits.h>
 # include <calc/byteswap.h>
 # include <calc/have_stdlib.h>
+# include <calc/attribute.h>
 #endif
 #ifdef HAVE_STDLIB_H
 # include <stdlib.h>
@@ -55,8 +57,7 @@
 
 
 #ifndef ALLOCTEST
-# define freeh(p) { if (((void *)p != (void *)_zeroval_) &&		\
-			 ((void *)p != (void *)_oneval_)) free((void *)p); }
+# define freeh(p) { if (!is_const(p)) { free((void *)(p)); } }
 #endif
 
 
@@ -150,7 +151,7 @@ typedef SB32 SFULL;			/* signed FULL */
 
 
 /*
- * other misc typedefs
+ * other miscellaneous typedefs
  */
 typedef USB32 QCKHASH;			/* 32 bit hash value */
 #if defined(HAVE_B64) && LONG_BITS == 32
@@ -289,6 +290,7 @@ typedef struct {
  * Function prototypes for integer math routines.
  */
 E_FUNC HALF * alloc(LEN len);
+E_FUNC int is_const(HALF* h);
 #ifdef	ALLOCTEST
 E_FUNC void freeh(HALF *);
 #endif
@@ -397,7 +399,7 @@ E_FUNC void zlcmfact(ZVALUE z, ZVALUE *dest);
 
 
 /*
- * Misc misc functions. :-)
+ * miscellaneous functions. :-)
  */
 E_FUNC void zsquaremod(ZVALUE z1, ZVALUE z2, ZVALUE *res);
 E_FUNC void zminmod(ZVALUE z1, ZVALUE z2, ZVALUE *res);
@@ -579,7 +581,12 @@ E_FUNC void zredcpower(REDC *rp, ZVALUE z1, ZVALUE z2, ZVALUE *res);
 #define zcopyval(z1,z2) memcpy((z2).v, (z1).v, (z1).len * sizeof(HALF))
 #define zquicktrim(z)	{if (((z).len > 1) && ((z).v[(z).len-1] == 0)) \
 				(z).len--;}
-#define zfree(z)	freeh((z).v)
+#define zfree(z)	{if ((z).len != 0 && (z).v != NULL) { \
+			    freeh((z).v); \
+			    (z).v = NULL; \
+			    (z).len = 0; \
+			    (z).sign = 0; } \
+			}
 
 
 /*
@@ -594,7 +601,8 @@ E_FUNC void zredcpower(REDC *rp, ZVALUE z1, ZVALUE z2, ZVALUE *res);
 #define MODE_OCTAL	6
 #define MODE_BINARY	7
 #define MODE_REAL_AUTO	8
-#define MODE_MAX	8
+#define MODE_ENG    9
+#define MODE_MAX	9
 #define MODE2_OFF	(MODE_MAX+1)
 
 /* XXX - perhaps we need the MODE_REAL_AUTO vs MODE_REAL as a config mode? */
@@ -619,13 +627,14 @@ E_FUNC char *math_getdivertedio(void);
 E_FUNC int math_setmode(int mode);
 E_FUNC int math_setmode2(int mode);
 E_FUNC LEN math_setdigits(LEN digits);
-E_FUNC void math_fmt(char *, ...) PRINTF_FORMAT(1, 2);
+E_FUNC void math_fmt(char *, ...) __attribute__((format(printf, 1, 2)));
 
 
 /*
  * The error routine.
  */
-E_FUNC void math_error(char *, ...) PRINTF_FORMAT(1, 2);
+E_FUNC void math_error(char *, ...) \
+	__attribute__((format(printf, 1, 2))) __attribute__((noreturn));
 
 
 /*
@@ -641,19 +650,35 @@ E_FUNC ZVALUE *swap_HALF_in_ZVALUE(ZVALUE *dest, ZVALUE *src, BOOL all);
 /*
  * constants used often by the arithmetic routines
  */
-EXTERN HALF _zeroval_[], _oneval_[], _twoval_[], _threeval_[], _fourval_[];
-EXTERN HALF _fiveval_[], _sixval_[], _sevenval_[], _eightval_[], _nineval_[];
-EXTERN HALF _tenval_[], _elevenval_[], _twelveval_[], _thirteenval_[];
-EXTERN HALF _fourteenval_[], _fifteenval_[];
+EXTERN HALF _zeroval_[];
+EXTERN ZVALUE _zero_;
+
+EXTERN HALF _oneval_[];
+EXTERN ZVALUE _one_;
+EXTERN ZVALUE _neg_one_;
+
+EXTERN HALF _twoval_[];
+EXTERN ZVALUE _two_;
+
+EXTERN HALF _tenval_[];
+EXTERN ZVALUE _ten_;
+
 EXTERN HALF _sqbaseval_[];
-EXTERN HALF _fourthbaseval_[];
+EXTERN ZVALUE _sqbase_;
 
-EXTERN ZVALUE zconst[];		/* ZVALUE integers from 0 thru 15 */
+EXTERN HALF _pow4baseval_[];
+EXTERN ZVALUE _pow4base_;
 
-EXTERN ZVALUE _zero_, _one_, _two_, _ten_, _neg_one_;
-EXTERN ZVALUE _sqbase_, _pow4base_, _pow8base_;
+EXTERN HALF _pow8baseval_[];
+EXTERN ZVALUE _pow8base_;
 
-EXTERN ZVALUE _b32_, _b64_;
+/* _b32_ is _sqbaseval_ or _pow4baseval_ depending on BASEB */
+EXTERN ZVALUE _b32_;
+
+/* _b64_ is _pow4baseval_ or _pow8baseval_ depending on BASEB */
+EXTERN ZVALUE _b64_;
+
+EXTERN HALF *half_tbl[]; /* preset HALF constants, NULL terminated list */
 
 EXTERN BOOL _math_abort_;	/* nonzero to abort calculations */
 EXTERN ZVALUE _tenpowers_[];	/* table of 10^2^n */
